/*
 * insane.c --  source for the "insane" module, a sample virtual interface
 *
 * Copyright (c) 2000 Alessandro Rubini (rubini@linux.it)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
#ifndef __KERNEL__
#  define __KERNEL__
#endif

/* The Makefile takes care of adding -DMODULE */

#include <linux/etherdevice.h>
#include <linux/module.h>

#include <linux/kernel.h>	/* printk() */
#include <linux/slab.h>		/* kmalloc() */
#include <linux/errno.h>	/* error codes */
#include <linux/netdevice.h>	/* basic data structures */
#include <linux/init.h>		/* __init */
#include <linux/if_arp.h>	/* ARPHRD_ETHER */
#include <linux/version.h>	/* LINUX_VERSION_CODE */
#include <net/arp.h>		/* neighbor stuff */
#include <asm/uaccess.h>	/* memcpy and such */

#include "insane.h"

#define DRIVER_AUTHOR		"Port: Filippo Bistaffa <liquidator87@gmail.com>"
				// Source: Alessandro Rubini <rubini@linux.it>
#define DRIVER_DESC		"Insane Driver"

MODULE_LICENSE("GPL");
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);

struct net_device *insane_dev;

/* --------------------------------------------------------------------------
 * definition of the "private" data structure used by this interface
 */
struct insane_private {
    struct net_device_stats priv_stats;
    struct net_device *priv_device; /* interface used to xmit data */
    int priv_mode; /* how to drop packets */
    int priv_arg1; /* arguments to the dropping mode */
    int priv_arg2;
};

/* --------------------------------------------------------------------------
 * open and close
 */
int insane_open(struct net_device *dev)
{
    /* mark the device as operational */
    printk("%s: device opened\n", dev->name);
    netif_start_queue(dev);
    return 0;
}

int insane_close(struct net_device *dev)
{
    printk("%s: device closed\n", dev->name);
    netif_stop_queue(dev);
    return 0;
}

/* --------------------------------------------------------------------------
 * get_stats: return a pointer to the device statistics
 */
struct net_device_stats *insane_get_stats(struct net_device *dev)
{
    return &((struct insane_private *)netdev_priv(dev))->priv_stats;
}

/* --------------------------------------------------------------------------
 * header stuff: fall back on the slave interface to deal with this stuff
 */
static int insane_create_header(struct sk_buff *skb, struct net_device *dev, 
        unsigned short type, const void *daddr, const void *saddr, unsigned len)
{
    struct insane_private *priv = netdev_priv(insane_dev);
    int retval;

    skb->dev = priv->priv_device;
    retval = skb->dev->header_ops->create(skb, skb->dev, type, 
					  daddr, saddr, len);
    skb->dev = dev;
    return retval;
}

static int insane_rebuild_header(struct sk_buff *skb)
{
    struct insane_private *priv = netdev_priv(insane_dev);
    int retval;

    skb->dev = priv->priv_device;
    retval = skb->dev->header_ops->rebuild(skb);
    skb->dev = insane_dev;
    return retval;
}

/* --------------------------------------------------------------------------
 * create default header_ops struct
 */
static const struct header_ops insane_header_ops = {
    .create  = insane_create_header,
    .rebuild = insane_rebuild_header,
    .cache   = NULL, /* disable caching */
};

/* --------------------------------------------------------------------------
 * neighbors: this comes form shaper.c (Alan Cox) and is needed for ARP to work
 */
int insane_neigh_setup(struct neighbour *n)
{
    if (n->nud_state == NUD_NONE) {
        n->ops = &arp_broken_ops;
        n->output = n->ops->output;
    }
    return 0;
}

int insane_neigh_setup_dev(struct net_device *dev, struct neigh_parms *p)
{
    if (p->tbl->family == AF_INET) {
        p->neigh_setup = insane_neigh_setup;
        p->ucast_probes = 0;
        p->mcast_probes = 0;
    }
    return 0;
}

/* --------------------------------------------------------------------------
 * xmit: actual delivery (or not) of the data packets
 */
int insane_xmit(struct sk_buff *skb, struct net_device *dev)
{
    struct insane_private *priv = netdev_priv(dev);
    int accept; /* accept this packet or drop it */
    static unsigned long randval;
    
    if (!priv->priv_device) { 
        /* cannot send to anyone, just return */
        kfree_skb(skb);
        priv->priv_stats.tx_errors++;
        priv->priv_stats.tx_dropped++;
        return 0;
    }
    switch (priv->priv_mode) {
    
        case INSANE_PERCENT:
            if (!randval) randval = jiffies; /* a typical seed */
            /* hash the value, according to the TYPE_0 rule of glibc */
            randval = ((randval * 1103515245) + 12345) & 0x7fffffff;
            accept = (randval % 100) < priv->priv_arg1;
            break;

        case INSANE_TIME:
            randval = jiffies % (priv->priv_arg1 + priv->priv_arg2);
            accept = randval < priv->priv_arg1;
            break;
        
        case INSANE_PASS:
        default: /* unknown mode: default to pass */
            accept = 1;
    }
    
    if (!accept) {
        kfree_skb(skb);
        priv->priv_stats.tx_errors++;
        priv->priv_stats.tx_dropped++;
        return 0;
    }
    /* else, pass it to the real interface */

    priv->priv_stats.tx_packets++;
    priv->priv_stats.tx_bytes += skb->len;

#if 0
    return priv->priv_device->hard_start_xmit(skb, priv->priv_device);
#else
    skb->dev = priv->priv_device;
    skb->priority = 1;
    dev_queue_xmit (skb);
    return 0;
#endif
}

/* --------------------------------------------------------------------------
 * ioctl: let user programs configure this interface
 */
int insane_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
    int err;
    
    struct net_device *slave;
    struct insane_private *priv = netdev_priv(dev);
    /* hold a local (kernel-space) copy of the configuration data */
    struct insane_userinfo info;
    /* and a pointer into user space as well */
    struct insane_userinfo *uptr = (struct insane_userinfo *)ifr->ifr_data;
    
    /* only authorized users can control the interface */
    if (cmd == SIOCINSANESETINFO && !capable(CAP_NET_ADMIN))
        return -EPERM;
    
    /* process the command */
    switch(cmd) {
        case SIOCINSANEGETINFO: /* return configuration to user space */
        
            /* interface name */
            memset(info.name, 0, INSANE_NAMELEN);
            if (priv->priv_device)
                strncpy(info.name, priv->priv_device->name, INSANE_NAMELEN-1);
        
            /* parameters */
            info.mode = priv->priv_mode;
            info.arg1 = priv->priv_arg1;
            info.arg2 = priv->priv_arg2;
            /* return the data structure to  user space */
            err = copy_to_user(uptr, &info, sizeof(info));
            if (err) return err;
            break;

        case SIOCINSANESETINFO:
        
            /* retrieve the data structure from user space */
            err = copy_from_user(&info, uptr, sizeof(info));
            if (err) return err;

            printk("name: %s, arg %i %i\n", info.name, info.arg1, info.arg2);

            /* interface name */
            slave = __dev_get_by_name(&init_net, info.name);
            if (!slave)
                return -ENODEV;
            if (slave->type != ARPHRD_ETHER && slave->type != ARPHRD_LOOPBACK)
                return -EINVAL;

            /* The interface is good, get hold of it */
            priv->priv_device = slave;
            if (slave->header_ops)
                dev->header_ops = &insane_header_ops;
            else
                dev->header_ops = NULL;

            /* also, and clone its IP, MAC and other information */
            memcpy(dev->dev_addr,  slave->dev_addr,  sizeof(slave->dev_addr));
            memcpy(dev->broadcast, slave->broadcast, sizeof(slave->broadcast));

            /* accept the parameters (no checks here) */
            priv->priv_mode = info.mode;
            priv->priv_arg1 = info.arg1;
            priv->priv_arg2 = info.arg2;

            break;

        default:
            return -EOPNOTSUPP;
    }
    return 0;
}

/* --------------------------------------------------------------------------
 * create default netdev_ops struct
 */
static const struct net_device_ops insane_net_device_ops = {
    .ndo_open        = insane_open,
    .ndo_stop        = insane_close,
    .ndo_do_ioctl    = insane_ioctl,
    .ndo_get_stats   = insane_get_stats,
    .ndo_start_xmit  = insane_xmit,
    .ndo_neigh_setup = insane_neigh_setup_dev,
};

/* --------------------------------------------------------------------------
 * The initialization function: it is used to assign fields in the structure
 */

/*
 * The __init attribute has no effect (by now) in modules; it is nonetheless
 * good practice to declare initialization functions as such, when working
 * in Linux kernel space.
 */
void insane_init(struct net_device *dev)
{
    ether_setup(dev); /* assign some of the fields as "generic ethernet" */
    memset(netdev_priv(dev), 0, sizeof(struct insane_private));
    
    dev->netdev_ops = &insane_net_device_ops;
    /*

    dev->open            = insane_open;
    dev->stop            = insane_close;
    dev->do_ioctl        = insane_ioctl;
    dev->get_stats       = insane_get_stats;
    dev->hard_start_xmit = insane_xmit;
    dev->neigh_setup     = insane_neigh_setup_dev;

    */

    /* Assign random MAC address */
    random_ether_addr(dev->dev_addr);
}

/* --------------------------------------------------------------------------
 * module entry points
 */

int __init insane_init_module(void)
{
    int err;    
    insane_dev = alloc_netdev(sizeof(struct insane_private), 
                              "insane", insane_init);
    
    if (!insane_dev)
        return -ENOMEM;
    
    if (( err = dev_alloc_name(insane_dev, insane_dev->name) )) {
        printk(KERN_WARNING "%s: allocate name, error %i\n", 
               insane_dev->name, err);
        return -EIO;
    }
    
    if (( err = register_netdev(insane_dev) )) {
        printk(KERN_WARNING "%s: can't register, error %i\n", 
               insane_dev->name, err);
        return -EIO;
    }
    
    printk("%s: device registered\n", insane_dev->name);
    return 0;
}

void __exit insane_exit_module(void)
{    
    unregister_netdev(insane_dev);
    printk("%s: device unregistered\n", insane_dev->name);
    free_netdev(insane_dev);
}

module_init(insane_init_module);
module_exit(insane_exit_module);
